/*

GMP_ext.h -- extension to the GMP float functions for the iRRAM library
 
Copyright (C) 2002 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include <stdlib.h>
#include <stdio.h>

#include "gmp.h"
#include "gmp-mparam.h"


#define ABS(x) ((x) >= 0 ? (x) : -(x))
#define MAX(h,i) ((h) > (i) ? (h) : (i))



#ifdef __cplusplus
extern "C" {
#endif


#define GMP_min -1000000000
#define GMP_max 1000000000

#define MaxFreeVars 1000

#define GMP_PREC(x) ((unsigned long int) (x)->_mp_prec * BITS_PER_MP_LIMB - BITS_PER_MP_LIMB)
//#define GMP_PREC(x) mpf_get_prec(x)

#define GMP_SET_PREC(x,n) { if ( GMP_PREC(x)!= (n) ) {ext_gmp_set_prec(x,n);}}
//#define GMP_SET_PREC(x,n) { if ( GMP_PREC(x)!= (n) ) {mpf_set_prec(x,n);}}


const int limb_size_in_bit=8* sizeof (mp_limb_t);
const int half_limb_size_in_bit=4* sizeof (mp_limb_t);
const unsigned long one_shift_limb_size_in_bit_1 = 1ul<<(8* sizeof (mp_limb_t)-1);
const unsigned long one_shift_half_limb_size_in_bit =1ul<<(4* sizeof (mp_limb_t));

typedef __mpf_struct  * ext_gmp_type;
typedef struct {unsigned long mantissa; long exponent; } ext_gmp_sizetype;

typedef mpz_ptr int_gmp_type;


void ext_gmp_initialize(void);

long ext_gmp_size(ext_gmp_type z);


ext_gmp_type ext_gmp_init(void);
void ext_gmp_free(ext_gmp_type z);
void ext_gmp_getsize(ext_gmp_type z,ext_gmp_sizetype* s);
void ext_gmp_set_prec (mpf_ptr x, unsigned long int prec_in_bits);
void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2);
void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2);


extern long ext_gmp_var_count;
extern ext_gmp_type gmp_FreeVars[];
extern long gmp_FreeVarCount;



void ext_gmp_add(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p);
void ext_gmp_sub(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p);
void ext_gmp_mul(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p);
void ext_gmp_div(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p);
void ext_gmp_abs(ext_gmp_type z1,ext_gmp_type z);
void ext_gmp_truncate(ext_gmp_type z1,ext_gmp_type z);

void ext_gmp_set_z(ext_gmp_type r, int_gmp_type i);

void ext_gmp_add_i(ext_gmp_type z1, long z2, ext_gmp_type z, long p);
void ext_gmp_sub_i(ext_gmp_type z1, long z2, ext_gmp_type z, long p);
void ext_gmp_i_sub(long z1, ext_gmp_type z2, ext_gmp_type z, long p);
void ext_gmp_mul_i(ext_gmp_type z1, long z2, ext_gmp_type z, long p);
void ext_gmp_div_i(ext_gmp_type z1, long z2, ext_gmp_type z, long p);
void ext_gmp_i_div(long z1, ext_gmp_type z2, ext_gmp_type z, long p);

void ext_gmp_writee(ext_gmp_type z,long w);
char* ext_gmp_swritee(ext_gmp_type z, long w);
void ext_gmp_write(ext_gmp_type z,long w);

void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2);
void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2);

void ext_gmp_sqrt(ext_gmp_type z1,ext_gmp_type z,long p);
void ext_gmp_shift(ext_gmp_type z1,ext_gmp_type z,long p);



inline void ext_gmp_set_prec (mpf_ptr x, unsigned long int prec_in_bits)
{
  mp_size_t prec;
  mp_size_t size = ABS (x->_mp_size);
 
  prec = (MAX (53, prec_in_bits) + 2 * __GMP_BITS_PER_MP_LIMB - 1)/BITS_PER_MP_LIMB;
  
  if (size != prec + 1) x->_mp_d = (mp_ptr) (__gmp_reallocate_func)
    (x->_mp_d,
     0 , (prec + 1) * BYTES_PER_MP_LIMB);
//     (x->_mp_prec + 1) * BYTES_PER_MP_LIMB, (prec + 1) * BYTES_PER_MP_LIMB);
  x->_mp_prec = prec;
 
  /* If the precision decreased, truncate the number.  */
  if (size > prec + 1)
    x->_mp_size = x->_mp_size >= 0 ? (prec + 1) : -(prec + 1);
}

inline ext_gmp_type ext_gmp_init()
{
  ext_gmp_var_count +=1;
  if (gmp_FreeVarCount>0) {
    gmp_FreeVarCount -=1;
    return gmp_FreeVars[gmp_FreeVarCount];
  } else {
  ext_gmp_type z = (ext_gmp_type) malloc(sizeof (__mpf_struct));
  mpf_init(z);
  return z;
  }
}

inline void ext_gmp_free(ext_gmp_type z)
{ 
  if ( gmp_FreeVarCount < MaxFreeVars ) {
    gmp_FreeVars[gmp_FreeVarCount]=z;
    gmp_FreeVarCount +=1;
  } else {
    mpf_clear(z); 
    free (z);
  }
  ext_gmp_var_count -=1;
}

inline void ext_gmp_getsize(ext_gmp_type z,ext_gmp_sizetype* s)
{
  if ( z->_mp_size==0 ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else if ( ((z->_mp_d)[abs(z->_mp_size)-1]==0) ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else {
    s->mantissa=(z->_mp_d)[abs(z->_mp_size)-1];
    s->exponent=(z->_mp_exp-1) * limb_size_in_bit;

    if ( s->mantissa < one_shift_half_limb_size_in_bit )
      { s->mantissa = s->mantissa <<half_limb_size_in_bit;
        s->exponent-=half_limb_size_in_bit;
        if ( abs(z->_mp_size)>1 ) 
          s->mantissa+= (z->_mp_d)[abs(z->_mp_size)-2]
                         >>half_limb_size_in_bit;
    }
    if (s->mantissa > one_shift_limb_size_in_bit_1) {
      s->mantissa=(s->mantissa>>1)+1;
      s->exponent+=1;
    }
    else 
    if (abs(z->_mp_size)>0) s->mantissa+=1;
  }
//  printf("get %d*2^%d\n",s->mantissa,s->exponent);
}

inline void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2)
{
  unsigned long q1=GMP_PREC(z1);
  if (GMP_PREC(z2)< q1) GMP_SET_PREC(z2,q1);
  mpf_set(z2,z1);
}

inline void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2)
{
  unsigned long q1=GMP_PREC(z1);
  *z2=ext_gmp_init();
  if (GMP_PREC(*z2)< q1) GMP_SET_PREC(*z2,q1);
  mpf_set(*z2,z1);
}

#ifdef __cplusplus
}
#endif

