/*

MPFR_ext.h -- extension to the MPFR float functions for the iRRAM library
 
Copyright (C) 2002/2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#ifndef MPFR_ext_h
#define MPFR_ext_h

#include <stdlib.h>
#include <stdio.h>

// it would be better to include gmp-mparam.h !
#define BITS_PER_MP_LIMB 32 

#include "gmp.h"
#include "mpfr.h"

#define ABSSIZE(x) ((x)->_mpfr_size & ((1<<30)-1))
#define SIZE(x) ((x)->_mpfr_size)
#define EXP(x) ((x)->_mpfr_exp)
#define MANT(x) ((x)->_mpfr_d)
#define MPFR_SIGN(x) (((x)->_mpfr_size >> 31) ? -1 : 1)
#define PREC(x) ((x)->_mpfr_prec)
#define NOTZERO(x) (MANT(x)[(PREC(x)-1)/BITS_PER_MP_LIMB])
// #define mpfr_sgn(x) ((NOTZERO(x)) ? MPFR_SIGN(x) : 0)     
                      
#define GMP_min -1000000000
#define GMP_max 1000000000

#define iRRAM_mpfr_rounding_mode GMP_RNDN

#ifdef __cplusplus
extern "C" {
#endif

typedef __mpfr_struct  * ext_mpfr_type;
typedef struct {unsigned long mantissa; long exponent; } ext_mpfr_sizetype;
typedef mpz_ptr  int_mpfr_type;

extern long ext_mpfr_var_count;
void ext_mpfr_remove_trailing_zeroes (mpfr_t x);

void ext_mpfr_initialize(void);

ext_mpfr_type ext_mpfr_init(void);
void ext_mpfr_free(ext_mpfr_type z);
long ext_mpfr_size(ext_mpfr_type z);
void ext_mpfr_getsize(ext_mpfr_type z,ext_mpfr_sizetype* s);

void ext_mpfr_add(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p);
void ext_mpfr_sub(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p);
void ext_mpfr_mul(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p);
void ext_mpfr_div(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p);
void ext_mpfr_abs(ext_mpfr_type z1,ext_mpfr_type z);
void ext_mpfr_truncate(ext_mpfr_type z1,ext_mpfr_type z);

void ext_mpfr_set_z(ext_mpfr_type r, int_mpfr_type i);
void ext_mpfr_set_f2i(ext_mpfr_type r, int_mpfr_type i);

void ext_mpfr_add_i(ext_mpfr_type z1, long z2, ext_mpfr_type z, long p);
void ext_mpfr_sub_i(ext_mpfr_type z1, long z2, ext_mpfr_type z, long p);
void ext_mpfr_i_sub(long z1, ext_mpfr_type z2, ext_mpfr_type z, long p);
void ext_mpfr_mul_i(ext_mpfr_type z1, long z2, ext_mpfr_type z, long p);
void ext_mpfr_div_i(ext_mpfr_type z1, long z2, ext_mpfr_type z, long p);
void ext_mpfr_i_div(long z1, ext_mpfr_type z2, ext_mpfr_type z, long p);

void ext_mpfr_writee(ext_mpfr_type z,long w);
void ext_mpfr_write(ext_mpfr_type z,long w);

void ext_mpfr_duplicate_w_init(ext_mpfr_type z1,ext_mpfr_type *z2);
void ext_mpfr_duplicate_wo_init(ext_mpfr_type z1,ext_mpfr_type z2);

void ext_mpfr_sqrt(ext_mpfr_type z1,ext_mpfr_type z,long p);
void ext_mpfr_shift(ext_mpfr_type z1,ext_mpfr_type z,long p);

#ifdef __cplusplus
}
#endif

#endif
