/*

iRRAM_INTEGER.h -- header file for the INTEGER part of the iRRAM library
 
Copyright (C) 2001-2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

class INTEGER 
{
public:

friend class REAL;
friend class DYADIC;
friend class RATIONAL;

/****** Constructors ******/

INTEGER(MP_int_type y);

INTEGER();
INTEGER(const int i);
INTEGER(const long i);
INTEGER(const char* s);
INTEGER(const INTEGER& y); 
INTEGER(const DYADIC& y); 
INTEGER(const REAL& y);  
INTEGER(const double d);

/****** Copy constructor ******/

INTEGER& operator = (const int y);
INTEGER& operator = (const long y);
INTEGER& operator = (const INTEGER& y);

	/****** Destructor ******/
	
~INTEGER();

	/****** Standard arithmetic ******/

friend INTEGER 	operator +  (const INTEGER& x, const INTEGER& y);
friend INTEGER 	operator +  (const INTEGER& x, const int      y);
friend INTEGER 	operator +  (const INTEGER& x, const long     y);
friend INTEGER 	operator +  (const int      x, const INTEGER& y);
friend INTEGER 	operator +  (const long     x, const INTEGER& y);

friend INTEGER 	operator -  (const INTEGER& x, const INTEGER& y);
friend INTEGER 	operator -  (const INTEGER& x, const int      y);
friend INTEGER 	operator -  (const INTEGER& x, const long     y);
friend INTEGER 	operator -  (const int      x, const INTEGER& y);
friend INTEGER 	operator -  (const long     x, const INTEGER& y);

friend INTEGER 	operator -  (const INTEGER& x);

friend INTEGER 	operator *  (const INTEGER& x, const INTEGER& y);
friend INTEGER	operator *  (const INTEGER& x, const int      y);
friend INTEGER	operator *  (const INTEGER& x, const long     y);
friend INTEGER	operator *  (const int      x, const INTEGER& y);
friend INTEGER	operator *  (const long     x, const INTEGER& y);

friend INTEGER 	operator /  (const INTEGER& x, const INTEGER& y);
friend INTEGER 	operator /  (const INTEGER& x, const int      y);
friend INTEGER 	operator /  (const INTEGER& x, const long     y);
friend INTEGER 	operator /  (const int      x, const INTEGER& y);
friend INTEGER 	operator /  (const long     x, const INTEGER& y);

friend INTEGER 	operator ^  (const INTEGER& x, const INTEGER& y);
friend INTEGER  operator ^  (const INTEGER& x, const unsigned long y);
friend INTEGER 	operator << (const INTEGER& x, const long y);
friend INTEGER 	operator >> (const INTEGER& x, const long y);
friend INTEGER	operator %  (const INTEGER& x, const INTEGER& y);

friend INTEGER&	operator += (	   INTEGER& x, const INTEGER& y);
friend INTEGER&	operator += (	   INTEGER& x, const long     y);
friend INTEGER&	operator += (	   INTEGER& x, const int      y);

friend INTEGER&	operator -= (	   INTEGER& x, const INTEGER& y);
friend INTEGER&	operator -= (	   INTEGER& x, const long     y);
friend INTEGER&	operator -= (	   INTEGER& x, const int      y);

friend INTEGER&	operator *= (	   INTEGER& x, const INTEGER& y);
friend INTEGER&	operator *= (	   INTEGER& x, const int      y);
friend INTEGER&	operator *= (	   INTEGER& x, const long     y);



friend INTEGER 	sqrt	(const INTEGER& x);
friend INTEGER	scale	(const INTEGER& x, const long k);
friend INTEGER	root    (const INTEGER& x, const unsigned long n);
friend INTEGER	abs	(const INTEGER& x);
friend INTEGER	log2	(const INTEGER& x);

/****** Comparisons ******/

friend bool operator <  (const INTEGER& x, const INTEGER& y);
friend bool operator <  (const INTEGER& x, const int      y);
friend bool operator <  (const INTEGER& x, const long     y);
friend bool operator <  (const int      x, const INTEGER& y);
friend bool operator <  (const long     x, const INTEGER& y);

friend bool operator <= (const INTEGER& x, const INTEGER& y);
friend bool operator <= (const INTEGER& x, const int      y);
friend bool operator <= (const INTEGER& x, const long     y);
friend bool operator <= (const int      x, const INTEGER& y);
friend bool operator <= (const long     x, const INTEGER& y);

friend bool operator >  (const INTEGER& x, const INTEGER& y);
friend bool operator >  (const INTEGER& x, const int      y);
friend bool operator >  (const INTEGER& x, const long     y);
friend bool operator >  (const int      x, const INTEGER& y);
friend bool operator >  (const long     x, const INTEGER& y);

friend bool operator >= (const INTEGER& x, const INTEGER& y);
friend bool operator >= (const INTEGER& x, const int      y);
friend bool operator >= (const INTEGER& x, const long     y);
friend bool operator >= (const int      x, const INTEGER& y);
friend bool operator >= (const long     x, const INTEGER& y);

friend bool operator == (const INTEGER& x, const INTEGER& y);
friend bool operator == (const INTEGER& x, const int      y);
friend bool operator == (const INTEGER& x, const long     y);
friend bool operator == (const int      x, const INTEGER& y);
friend bool operator == (const long     x, const INTEGER& y);

friend bool operator != (const INTEGER& x, const INTEGER& y);
friend bool operator != (const INTEGER& x, const int      y);
friend bool operator != (const INTEGER& x, const long     y);
friend bool operator != (const int      x, const INTEGER& y);
friend bool operator != (const long     x, const INTEGER& y);
	
/****** Output ******/
friend void 	rwrite	(const INTEGER& x, const long w);
friend void     rwrite  (const INTEGER& x);

friend char*    swrite  (const INTEGER& x, const long w);
friend char*    swrite  (const INTEGER& x);

/* conversion */
operator int()  const ;
operator long() const ;

/****** Implementational issues ******/

int size(); //****** will be private

/****** Private ******/
private:

MP_int_type value;

};


