/*
/*

iRRAM.h -- header file for the REAL class of the iRRAM library
 
Copyright (C) 2001-2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

class REAL : public  METRIC_OBJECT
{
public:

// Constructors: -------------------------------

REAL();
REAL(int  i);
REAL(long i);
REAL(char* s);
REAL(const DYADIC& y);
REAL(const REAL& y);
REAL(double d);
REAL(const INTEGER& y);
REAL(const RATIONAL& y);

// Copy Constructor: ---------------------------

REAL& operator = (const REAL& y);

// Destructor: ---------------------------------

~REAL();

// Standard Arithmetic: ------------------------

friend REAL  operator +  (const REAL& x,	const REAL& y);
friend REAL  operator +  (const REAL& x,	long int n);
friend REAL  operator +  (long int n,		const REAL& x);
friend REAL  operator +  (const REAL& x,	int n);
friend REAL  operator +  (int n,		const REAL& x);
friend REAL& operator += (REAL& x,		const REAL& y);
			  
friend REAL  operator -  (const REAL& x, 	const REAL& y);
friend REAL  operator -  (const REAL& x,	long int n);
friend REAL  operator -  (long int n,		const REAL& x);
friend REAL  operator -  (const REAL& x,	int n);
friend REAL  operator -  (int n,		const REAL& x);
friend REAL  operator -  (const REAL& x);

friend REAL  operator *  (const REAL& x,	const REAL& y);
friend REAL  operator *  (const REAL& x, 	long int n);
friend REAL  operator *  (long int n, 		const REAL& x);
friend REAL  operator *  (const REAL& x, 	int n);
friend REAL  operator *  (int n, 		const REAL& x);			  
friend REAL& operator *= (REAL& x,		const REAL& y);
friend REAL& operator *= (REAL& x,		long int n);
			  
friend REAL  operator /  (const REAL& x,	const REAL& y);
friend REAL  operator /  (const REAL& x,	long int n);
friend REAL  operator /  (long int n,		const REAL& x);
friend REAL  operator /  (const REAL& x,	int n);
friend REAL  operator /  (int n,		const REAL& x);

friend REAL  operator ^  (const REAL& x,	const REAL& y);
friend REAL  operator ^  (const REAL& x,	long n);

friend REAL  operator << (const REAL& x,	long n);
friend REAL  operator >> (const REAL& x,	long n);

friend REAL  sqrt        (const REAL& x);

friend REAL  square      (const REAL& x);

friend REAL  scale       (const REAL& x,
                          long k);

// Comparisons: --------------------------------

friend LAZY_BOOLEAN  operator <  (const REAL&    x, const REAL&    y);
friend LAZY_BOOLEAN  operator <  (const REAL&    x, const int      y);
friend LAZY_BOOLEAN  operator <  (const REAL&    x, const long     y);
friend LAZY_BOOLEAN  operator <  (const REAL&    x, const double   y);
friend LAZY_BOOLEAN  operator <  (const REAL&    x, const INTEGER& y);
friend LAZY_BOOLEAN  operator <  (const REAL&    x, const DYADIC&  y);
friend LAZY_BOOLEAN  operator <  (const int      x, const REAL&    y);
friend LAZY_BOOLEAN  operator <  (const long     x, const REAL&    y);
friend LAZY_BOOLEAN  operator <  (const double   x, const REAL&    y);
friend LAZY_BOOLEAN  operator <  (const INTEGER& x, const REAL&    y);
friend LAZY_BOOLEAN  operator <  (const DYADIC&  x, const REAL&    y);

friend LAZY_BOOLEAN  operator <= (const REAL&    x, const REAL&    y);
friend LAZY_BOOLEAN  operator <= (const REAL&    x, const int      y);
friend LAZY_BOOLEAN  operator <= (const REAL&    x, const long     y);
friend LAZY_BOOLEAN  operator <= (const REAL&    x, const double   y);
friend LAZY_BOOLEAN  operator <= (const REAL&    x, const INTEGER& y);
friend LAZY_BOOLEAN  operator <= (const REAL&    x, const DYADIC&  y);
friend LAZY_BOOLEAN  operator <= (const int      x, const REAL&    y);
friend LAZY_BOOLEAN  operator <= (const long     x, const REAL&    y);
friend LAZY_BOOLEAN  operator <= (const double   x, const REAL&    y);
friend LAZY_BOOLEAN  operator <= (const INTEGER& x, const REAL&    y);
friend LAZY_BOOLEAN  operator <= (const DYADIC&  x, const REAL&    y);

friend LAZY_BOOLEAN  operator >  (const REAL&    x, const REAL&    y);
friend LAZY_BOOLEAN  operator >  (const REAL&    x, const int      y);
friend LAZY_BOOLEAN  operator >  (const REAL&    x, const long     y);
friend LAZY_BOOLEAN  operator >  (const REAL&    x, const double   y);
friend LAZY_BOOLEAN  operator >  (const REAL&    x, const INTEGER& y);
friend LAZY_BOOLEAN  operator >  (const REAL&    x, const DYADIC&  y);
friend LAZY_BOOLEAN  operator >  (const int      x, const REAL&    y);
friend LAZY_BOOLEAN  operator >  (const long     x, const REAL&    y);
friend LAZY_BOOLEAN  operator >  (const double   x, const REAL&    y);
friend LAZY_BOOLEAN  operator >  (const INTEGER& x, const REAL&    y);
friend LAZY_BOOLEAN  operator >  (const DYADIC&  x, const REAL&    y);

friend LAZY_BOOLEAN  operator >= (const REAL&    x, const REAL&    y);
friend LAZY_BOOLEAN  operator >= (const REAL&    x, const int      y);
friend LAZY_BOOLEAN  operator >= (const REAL&    x, const long     y);
friend LAZY_BOOLEAN  operator >= (const REAL&    x, const double   y);
friend LAZY_BOOLEAN  operator >= (const REAL&    x, const INTEGER& y);
friend LAZY_BOOLEAN  operator >= (const REAL&    x, const DYADIC&  y);
friend LAZY_BOOLEAN  operator >= (const int      x, const REAL&    y);
friend LAZY_BOOLEAN  operator >= (const long     x, const REAL&    y);
friend LAZY_BOOLEAN  operator >= (const double   x, const REAL&    y);
friend LAZY_BOOLEAN  operator >= (const INTEGER& x, const REAL&    y);
friend LAZY_BOOLEAN  operator >= (const DYADIC&  x, const REAL&    y);

friend LAZY_BOOLEAN  operator == (const REAL&    x, const REAL&    y);
friend LAZY_BOOLEAN  operator == (const REAL&    x, const int      y);
friend LAZY_BOOLEAN  operator == (const REAL&    x, const long     y);
friend LAZY_BOOLEAN  operator == (const REAL&    x, const double   y);
friend LAZY_BOOLEAN  operator == (const REAL&    x, const INTEGER& y);
friend LAZY_BOOLEAN  operator == (const REAL&    x, const DYADIC&  y);
friend LAZY_BOOLEAN  operator == (const int      x, const REAL&    y);
friend LAZY_BOOLEAN  operator == (const long     x, const REAL&    y);
friend LAZY_BOOLEAN  operator == (const double   x, const REAL&    y);
friend LAZY_BOOLEAN  operator == (const INTEGER& x, const REAL&    y);
friend LAZY_BOOLEAN  operator == (const DYADIC&  x, const REAL&    y);

friend LAZY_BOOLEAN  operator != (const REAL&    x, const REAL&    y);
friend LAZY_BOOLEAN  operator != (const REAL&    x, const int      y);
friend LAZY_BOOLEAN  operator != (const REAL&    x, const long     y);
friend LAZY_BOOLEAN  operator != (const REAL&    x, const double   y);
friend LAZY_BOOLEAN  operator != (const REAL&    x, const INTEGER& y);
friend LAZY_BOOLEAN  operator != (const REAL&    x, const DYADIC&  y);
friend LAZY_BOOLEAN  operator != (const int      x, const REAL&    y);
friend LAZY_BOOLEAN  operator != (const long     x, const REAL&    y);
friend LAZY_BOOLEAN  operator != (const double   x, const REAL&    y);
friend LAZY_BOOLEAN  operator != (const INTEGER& x, const REAL&    y);
friend LAZY_BOOLEAN  operator != (const DYADIC&  x, const REAL&    y);


friend LAZY_BOOLEAN  positive    (const REAL& x, const long k);

friend LAZY_BOOLEAN  bound       (const REAL& x, const long k);

friend DYADIC approx     (const REAL& x, const long p);

// Output: -------------------------------------

friend void   rwrite     (const REAL& x, const long p, const long w);
friend void   rwrite     (const REAL& x, const long w);
friend void   rwritee    (const REAL& x, const long w);
friend void   rshow      (const REAL& x, const long w);

void REAL::rcheck() const;

friend long    upperbound (const REAL& x);
friend long    size       (const REAL& x); 



friend REAL    round2     (const REAL& x); 
friend long    round      (const REAL& x); 

friend REAL    abs        (const REAL& x);
 

// limit operators: ------------------------

//friend REAL    limit      (REAL f(long, const REAL&),
//                           const REAL& x);

friend REAL    limit      (REAL f(long, const REAL&, const REAL&),
                           const REAL& x, 
                           const REAL& y);

friend REAL    limit      (REAL f(long));

friend REAL    limit_hint (REAL f(long, const REAL&),
                           long hint,
                           const REAL& x);

friend REAL    limit_hint (REAL f(long, const REAL&, const REAL&),
                           long hint,
                           const REAL& x, 
                           const REAL& y);

friend REAL    limit_hint (REAL f(long),
                           long hint);

friend REAL    limit_lip  (REAL f(long, const REAL&),
                           long lip,
                           bool on_domain(const REAL&),
                           const REAL& x);

friend REAL    limit_lip  (REAL f(long, const REAL&),
                           long lip(const REAL&),
                           const REAL& x);

friend REAL    limit_lip   (REAL f(long, const REAL&, const REAL&),
                           long lip,
                           bool on_domain(const REAL&,const REAL&),
                           const REAL& x,
                           const REAL& y);

friend REAL    iterate     (REAL (*)(REAL&, REAL&, const REAL&),
                           const REAL&, const REAL&, const REAL&);

friend REAL    iterate     (REAL (*)(REAL&, REAL&),
                           const REAL&, const REAL&);

// reduced error propagation: ------------------------

friend REAL    lipschitz  (REAL f(const REAL&),
                           long lip,
                           bool on_domain(const REAL&),
                           const REAL& x);

friend REAL    lipschitz  (REAL f(const REAL&),
                           REAL lip_f(const REAL&),
                           bool on_domain(const REAL&),
                           const REAL& x);

friend REAL    lipschitz  (REAL f(const REAL&,const REAL&),
                           long lip,
                           bool on_domain(const REAL&,const REAL&),
                           const REAL& x,
                           const REAL& y);

friend REAL    lipschitz  (REAL f(long, const REAL&),
                           long lip,
                           bool on_domain(long k,const REAL&),
                           long k,
                           const REAL& x);

friend REAL    lipschitz  (REAL f(long, const REAL&,const REAL&),
                           long lip,
                           bool on_domain(long k,const REAL&,const REAL&),
                           long k,
                           const REAL& x,
                           const REAL& y);
			   
friend REAL    lipschitz (REAL f(const REAL&),
                          REAL lip_f(const REAL&),
                          const REAL& x);

// coexistence with other classes: -------------

friend class REALMATRIX;
friend class SPARSEREALMATRIX;
friend class INTEGER;
friend class DYADIC;
friend class RATIONAL;

operator double() const;
operator long() const;

// for the sake of prooving computational adequacy:
// if q=module(f,x,p), then |z-x|<2^q => |f(z)-f(x)| < 2^p
friend long module(REAL f(const REAL&),const REAL& x, long p);


// implementational issues: --------------------

private:

MP_type       value; 
sizetype      error;


public:
REAL(MP_type y, const sizetype errorinfo);
void adderror (sizetype error);
void seterror (sizetype error);
void geterror (sizetype& error) const;

};

