/*

iRRAM_core.h -- basic file for the errorsize arithmetic 
                and for the templates in the iRRAM library
 
Copyright (C) 2001-2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/
#include <math.h>
#include "stdio.h"
#define do_not_activate_iRRAM_control
#include "iRRAM.h"
#include "iRRAM_version.h"


#ifndef IRRAM_CORE_H
#define IRRAM_CORE_H


inline long min(long a,long b)        
   { return (a < b ? a : b); };
inline long max(long a,long b)
   { return (a < b ? b : a); };
inline long max3(long a,long b,long c)
   { return max(a,max(b,c)); };
inline long max4(long a,long b,long c,long d)
   { return max(max(a,b),max(c,d)); };

extern int iRRAM_debug;
extern int iRRAM_infinite;
extern void resources(double&,unsigned long&);

class ITERATION_STACK;
extern ITERATION_STACK ACTUAL_STACK;

class ITERATION_STACK {
public:
  ITERATION_STACK();
  ~ITERATION_STACK();
  long prec_diff;
  long prec_policy;
  long prec_inc;
  double prec_factor;
  long inlimit;
  long actual_prec; 
};

extern long printfausgaben;
extern long printfaufrufe;

extern char * eingabe;
extern char * position;
extern long eingabeplatz;

extern unsigned long cachesize_test;
extern long * testwert;
extern unsigned long testanzahl;
extern unsigned long testaufrufe;

extern unsigned long cachesize_rkonvert;
extern MP_type * rkonvert;
extern unsigned long rkonvertanzahl;
extern unsigned long rkonvertaufrufe;

extern unsigned long cachesize_ikonvert;
extern MP_int_type * ikonvert;
extern unsigned long ikonvertanzahl;
extern unsigned long ikonvertaufrufe;

extern int iRRAM_debug;
extern long iRRAM_starting_prec;
extern long iRRAM_prec_inc0;
extern long iRRAM_prec_inc1;
extern long iRRAM_prec_inc2;


#ifndef NODEBUG
  #define DEBUG0(level,p) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) p
  #define DEBUG1(level,p) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) fprintf(stderr,p)
  #define DEBUG2(level,p1,p2) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) fprintf(stderr,p1,p2)
  #define DEBUG3(level,p1,p2,p3) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) fprintf(stderr,p1,p2,p3)
  #define DEBUG4(level,p1,p2,p3,p4) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) fprintf(stderr,p1,p2,p3,p4)
  #define DEBUG5(level,p1,p2,p3,p4,p5) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) fprintf(stderr,p1,p2,p3,p4,p5)
  #define DEBUG6(level,p1,p2,p3,p4,p5,p6) if (iRRAM_debug>=ACTUAL_STACK.inlimit+level) fprintf(stderr,p1,p2,p3,p4,p5,p6)
#else
  #define DEBUG0(level,p)
  #define DEBUG1(level,p) 
  #define DEBUG2(level,p1,p2) 
  #define DEBUG3(level,p1,p2,p3) 
  #define DEBUG4(level,p1,p2,p3,p4) 
  #define DEBUG5(level,p1,p2,p3,p4,p5) 
  #define DEBUG6(level,p1,p2,p3,p4,p5,p6)
#endif


inline void cacheinc_test(){
  testanzahl++;
  if ( testanzahl < cachesize_test ) return;
  if ( cachesize_test> 1000000000 ) 
       throw iRRAM_Numerical_Exception(iRRAM_cacheerror_test);
  long * temp = testwert;
  testwert = new long[2*cachesize_test];
  for (unsigned long i=0; i< cachesize_test;i++) testwert[i]=temp[i];
  cachesize_test=2*cachesize_test;
  delete []temp;
}

inline void cacheinc_rkonvert(){
  rkonvertanzahl++;
  if ( rkonvertanzahl < cachesize_rkonvert ) return;
  if ( cachesize_rkonvert> 1000000000 ) 
       throw iRRAM_Numerical_Exception(iRRAM_cacheerror_rkonvert);
  MP_type * temp = rkonvert;
  rkonvert = new MP_type[2*cachesize_rkonvert];
  for (unsigned long i=0; i< cachesize_rkonvert;i++) rkonvert[i]=temp[i];
  cachesize_rkonvert=2*cachesize_rkonvert;
  delete []temp;
}

inline void cacheinc_ikonvert(){
  ikonvertanzahl++;
  if ( ikonvertanzahl < cachesize_ikonvert ) return;
  if ( cachesize_ikonvert> 1000000000 ) 
       throw iRRAM_Numerical_Exception(iRRAM_cacheerror_ikonvert);
  MP_int_type * temp = ikonvert;
  ikonvert = new MP_int_type[2*cachesize_ikonvert];
  for (unsigned long i=0; i< cachesize_ikonvert;i++) ikonvert[i]=temp[i];
  cachesize_ikonvert=2*cachesize_ikonvert;
  delete []temp;
}











#define GUARD_BITS  ( (long) (8*sizeof(SIZETYPEMANTISSA)-3) )

struct Iteration {long prec_diff; Iteration (long p){prec_diff=p;}; };

// inline void REITERATE(long p_diff){  throw Iteration(p_diff); }
#define REITERATE(x) throw Iteration(x)

inline long scale(long w,long p) {return ((p>GUARD_BITS)?0:(w>>p)); };





const SIZETYPEMANTISSA max_mantissa=((long)1) <<  GUARD_BITS   ; 
const SIZETYPEMANTISSA min_mantissa= ( 1L << (GUARD_BITS-8) );
const SIZETYPEEXPONENT  min_exponent=MP_min+8*sizeof(SIZETYPEEXPONENT);

inline void sizetype_normalize( sizetype& e) {
  if ( e.mantissa >=  max_mantissa ) {
      e.mantissa =  (e.mantissa>>2) +1; 
      e.exponent += 2;
  } else if ( e.mantissa < min_mantissa) {
      e.mantissa =  e.mantissa<<6;
      e.exponent -= 6;
  }
  if ( e.exponent < MP_min ) {
    e.exponent = min_exponent;
    e.mantissa = 1;
  }
}

inline void sizetype_inc(sizetype& x,const sizetype& y) 
{ SIZETYPEEXPONENT exponent;
  exponent=max(x.exponent,y.exponent);
  x.mantissa=scale(x.mantissa,(exponent-x.exponent))
            +scale(y.mantissa,(exponent-y.exponent)) +1 ;
  x.exponent=exponent;
  sizetype_normalize(x);
}

inline void sizetype_dec(sizetype& x, const sizetype& y ) 
{ x.mantissa=x.mantissa - scale(y.mantissa,(x.exponent-y.exponent)) -1 ;
  sizetype_normalize(x);
}
inline void sizetype_dec(sizetype& x) 
{ x.mantissa -= 1 ;
}

inline void sizetype_add(sizetype& x,const sizetype& y,const sizetype& z) 
{
  x.exponent=max(y.exponent,z.exponent);
  x.mantissa=scale(z.mantissa,(x.exponent-z.exponent))
            +scale(y.mantissa,(x.exponent-y.exponent)) +1 ;
  sizetype_normalize(x);
}

inline void sizetype_copy(sizetype& x,const sizetype& y) 
{
  x.exponent=y.exponent;
  x.mantissa=y.mantissa;
}

inline void sizetype_half(sizetype& x,const sizetype& y) 
{
  x.exponent=y.exponent-1;
  x.mantissa=y.mantissa;
}


inline void sizetype_sqrt(sizetype& x,const sizetype& y) 
{
  if ((y.exponent&1) == 1) {
    x.exponent=(y.exponent-1)/2; 
    x.mantissa=y.mantissa <<1;
  }  else {
    x.exponent=y.exponent/2; 
    x.mantissa=y.mantissa;
  }
  x.mantissa=long(sqrt(double(x.mantissa)))+1;
}

inline void sizetype_shift(sizetype& x,const sizetype& y,const long s) 
{
  x.exponent=y.exponent+s;
  x.mantissa=y.mantissa;
}

inline void sizetype_mult(sizetype& x,const sizetype& y,const sizetype& z) 
{ unsigned long long lmantissa=
     ((unsigned long long)(y.mantissa))*z.mantissa;
  x.exponent=y.exponent+z.exponent;
  while (lmantissa >= (1<<GUARD_BITS) ) 
       { lmantissa=lmantissa>>4;x.exponent+=4;}
  x.mantissa=lmantissa+1;
  sizetype_normalize(x);
}

inline void sizetype_div(sizetype& x,const sizetype& y,const sizetype& z) 
{ unsigned long long lmantissa=
         (((unsigned long long)(y.mantissa))<<GUARD_BITS)/z.mantissa;
  x.exponent=y.exponent-z.exponent-GUARD_BITS;
  while (lmantissa >= (1<<GUARD_BITS) )
      { lmantissa=lmantissa>>4;x.exponent+=4; }
  x.mantissa=lmantissa+1;
  sizetype_normalize(x);
}

/*Compute maximum of y and z */
inline void sizetype_max(sizetype& x,const sizetype& y,const sizetype& z) 
{ if (y.exponent>z.exponent) 
  {
    x.mantissa=scale(z.mantissa,y.exponent-z.exponent);
    if (x.mantissa>=y.mantissa) x=z; else x=y;
  } else {
    x.mantissa=scale(y.mantissa,z.exponent-y.exponent);
    if (x.mantissa>=z.mantissa) x=y; else x=z;
  }
}

/*Test whether  y<z; for y=z the result might be wrong  */
inline int sizetype_less(const sizetype& y,const sizetype& z) 
{ unsigned long mantissa;
  if (y.exponent>z.exponent) 
  {
    mantissa=scale(z.mantissa,y.exponent-z.exponent);
    return (mantissa>=y.mantissa);
  } else {
    mantissa=scale(y.mantissa,z.exponent-y.exponent);
    return (mantissa<z.mantissa);
  }
}

inline void sizetype_set(sizetype& x,const long mantissa,const long exponent) 
{
  x.exponent=exponent;
  x.mantissa=mantissa;
  sizetype_normalize(x);
}

inline void sizetype_exact(sizetype& x) 
{
  x.exponent=MP_min;
  x.mantissa=0;
}


inline void ADJUST( sizetype& e) {
  if ( e.exponent < MP_min ) printf ("Warning, underflow in exponent!\n");
  if ( e.exponent >= MP_max ) 
  {
    DEBUG1(1,"actual precision infinite, ");
    REITERATE(0);
  }
}




/*****************************************/
// iRRAM_exec template

template <class ARGUMENT, class RESULT> RESULT iRRAM_exec 
(RESULT iRRAM_compute(ARGUMENT), ARGUMENT x)
{
RESULT result;

position=eingabe;
eingabe[0]=0;
printfausgaben=0;
printfaufrufe=0;
testaufrufe=0;
testanzahl=0;
rkonvertanzahl=0;ikonvertanzahl=0;
rkonvertaufrufe=0;ikonvertaufrufe=0;

ITERATION_STACK SAVED_STACK;

ACTUAL_STACK.prec_inc=iRRAM_prec_inc0;
ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

DEBUG3(1,"\niRRAM (version %s, backend %s) starting...\n",iRRAM_VERSION_rt,iRRAM_BACKEND);

while (true) {
  long p_end;
  try { result=iRRAM_compute(x);   if (!iRRAM_infinite) break;}
  catch ( Iteration it)  { p_end=ACTUAL_STACK.actual_prec+it.prec_diff; }
  catch ( const iRRAM_Numerical_Exception exc)
    {
      fprintf(stderr,"iRRAM exception: %s\n",iRRAM_error_msg[exc.type]);
      throw iRRAM_Numerical_Exception (exc);
    }

    	do {
      		ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)
		                         +iRRAM_prec_inc1;
      		ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
    	} while ( ACTUAL_STACK.actual_prec > p_end ); 
    	ACTUAL_STACK.inlimit=0;
if (iRRAM_debug>0) {
    	DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
	DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 
#ifdef	MP_space_count
    	DEBUG2(1,"   MP-memory in use:   %ld kB\n",MP_space_count); 
    	DEBUG2(1,"   max MP-memory used: %ld kB\n",MP_max_space_count); 
#endif
    	DEBUG2(1,"   cached tests:       %ld \n",testanzahl); 
	DEBUG2(1,"   cached conversions: %ld \n",rkonvertanzahl+ikonvertanzahl); 
	DEBUG2(1,"   cached inputs:      %d \n",position-eingabe); 
	DEBUG2(1,"   number of outputs:  %ld \n",printfaufrufe);
double time;
unsigned long memory;
resources(time,memory);
fprintf(stderr,"   accumul. CPU-Time:  %9e s\n",time);
fprintf(stderr,"   used Memory:        %9ld KB\n",memory/1024);
	DEBUG2(1,"increasing precision bound to 2^(%ld)...\n",ACTUAL_STACK.actual_prec);
    	}
	position=eingabe;
	printfaufrufe=0;
	testaufrufe=0;
  }


if (iRRAM_debug>0) {
DEBUG1(1,"iRRAM ending \n");
DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 
#ifdef	MP_space_count
DEBUG2(1,"   MP-memory in use:   %ld kB\n",MP_space_count); 
DEBUG2(1,"   max MP-memory used: %ld kB\n",MP_max_space_count); 
#endif
DEBUG2(1,"   cached tests:       %ld \n",testanzahl); 
DEBUG2(1,"   cached conversions: %ld \n",rkonvertanzahl+ikonvertanzahl); 
DEBUG2(1,"   cached inputs:      %d \n",position-eingabe); 
DEBUG2(1,"   number of outputs:  %ld \n",printfaufrufe); 
double time;
unsigned long memory;
resources(time,memory);
fprintf(stderr,"   total CPU-Time:     %9e s\n",time);
fprintf(stderr,"   total Memory:       %9ld KB\n",memory/1024);

DEBUG2(1,"   maximal precision:  %ld \n",ACTUAL_STACK.actual_prec); 
DEBUG2(1,"   maximal precision:  %ld \n",ACTUAL_STACK.actual_prec); 
}

for (long i= 0; i < rkonvertanzahl; i++) {
  MP_clear(rkonvert[i]);
  rkonvert[i]=0;
  }

for (long i= 0; i < ikonvertanzahl; i++) {
  MP_int_clear(ikonvert[i]);
  ikonvert[i]=0;
  }

rkonvertanzahl=0;
ikonvertanzahl=0;
DEBUG1(1,"\n   after cleaning up:\n");
DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 

return result;
};

template <class ARGUMENTA, class ARGUMENTB, class RESULT> RESULT iRRAM_exec 
(RESULT iRRAM_compute(ARGUMENTA, ARGUMENTB), ARGUMENTA x, ARGUMENTB y)
{
RESULT result;

position=eingabe;
eingabe[0]=0;
printfausgaben=0;
printfaufrufe=0;
testaufrufe=0;
testanzahl=0;
ikonvertanzahl=0;rkonvertanzahl=0;
ikonvertaufrufe=0;rkonvertaufrufe=0;

ITERATION_STACK SAVED_STACK;

ACTUAL_STACK.prec_inc=iRRAM_prec_inc0;
ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

DEBUG3(1,"\niRRAM (version %s, backend %s) starting...\n",iRRAM_VERSION_rt,iRRAM_BACKEND);

while (true) {
  long p_end;
  try { result=iRRAM_compute(x,y);   if (! iRRAM_infinite) break;}
  catch ( Iteration it)  { p_end=ACTUAL_STACK.actual_prec+it.prec_diff; }
  catch (const iRRAM_Numerical_Exception exc)
    {
      fprintf(stderr,"iRRAM exception: %s\n",iRRAM_error_msg[exc.type]);
      throw iRRAM_Numerical_Exception (exc);
    }
    	do {
      		ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc *
					ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      		ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc;
    	} while ( ACTUAL_STACK.actual_prec > p_end ); 
    	ACTUAL_STACK.inlimit=0;
if (iRRAM_debug>0) {
    	DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
	DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 
#ifdef	MP_space_count
    	DEBUG2(1,"   MP-memory in use:   %ld kB\n",MP_space_count); 
    	DEBUG2(1,"   max MP-memory used: %ld kB\n",MP_max_space_count); 
#endif
    	DEBUG2(1,"   cached tests:       %ld \n",testanzahl); 
	DEBUG2(1,"   cached conversions: %ld \n",rkonvertanzahl+ikonvertanzahl); 
	DEBUG2(1,"   cached inputs:      %d \n",position-eingabe); 
	DEBUG2(1,"   number of outputs:  %ld \n",printfaufrufe);
double time;
unsigned long memory;
resources(time,memory);
fprintf(stderr,"   accumul. CPU-Time:  %9e s\n",time);
fprintf(stderr,"   used Memory:        %9ld KB\n",memory/1024);
	DEBUG2(1,"increasing precision bound to 2^(%ld)...\n",ACTUAL_STACK.actual_prec);
        }
    	position=eingabe;
	printfaufrufe=0;
	testaufrufe=0;
  }


if (iRRAM_debug>0) {
DEBUG1(1,"iRRAM ending \n");
DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 
#ifdef	MP_space_count
DEBUG2(1,"   MP-memory in use:   %ld kB\n",MP_space_count); 
DEBUG2(1,"   max MP-memory used: %ld kB\n",MP_max_space_count); 
#endif
DEBUG2(1,"   cached tests:       %ld \n",testanzahl); 
DEBUG2(1,"   cached conversions: %ld \n",rkonvertanzahl+ikonvertanzahl); 
DEBUG2(1,"   cached inputs:      %d \n",position-eingabe); 
DEBUG2(1,"   number of outputs:  %ld \n",printfaufrufe); 
double time;
unsigned long memory;
resources(time,memory);
fprintf(stderr,"   total CPU-Time:     %9e s\n",time);
fprintf(stderr,"   total Memory:       %9ld KB\n",memory/1024);
DEBUG2(1,"   maximal precision:  %ld \n",ACTUAL_STACK.actual_prec); 
}

for (long i= 0; i < rkonvertanzahl; i++) {
  MP_clear(rkonvert[i]);
  rkonvert[i]=0;
  }

for (long i= 0; i < ikonvertanzahl; i++) {
  MP_int_clear(ikonvert[i]);
  ikonvert[i]=0;
  }

rkonvertanzahl=0;
ikonvertanzahl=0;
DEBUG1(1,"\n   after cleaning up:\n");
DEBUG2(1,"   objects in use:     %ld\n",METRIC_OBJECT_COUNT); 
DEBUG2(1,"   MP-objects in use:  %ld\n",MP_var_count); 

return result;
};


//*************************************************************************************/
// predefined instantiations of some iRRAM_exec routines can be found in stack.cc
//*************************************************************************************/
#endif /* ! iRRAM_CORE_H  */
