#include "gmp.h"
#include "gmp-impl.h"
#include "GMP_ext.h"
#include "longlong.h"

#define   min(a,b)        (a < b ? a : b)
#define   max(a,b)        (a < b ? b : a)
#define   max3(a,b,c)       max(a,max(b,c))
#define   max4(a,b,c,d)       max(max(a,b),max(c,d))

#ifndef DEBUG
#define DEBUG 0
#endif
#if DEBUG > 0
  long DEBUG_OP = 1;
  unsigned long op_count=0;
  long minshowop = DEBUG ;
#endif

long ext_gmp_var_count=0;

#define MaxFreeVars 1000
ext_gmp_type gmp_FreeVars[MaxFreeVars];
long gmp_FreeVarCount=0L;

void ext_gmp_initialize()
{
mpf_set_default_prec(53);
}

inline long ext_gmp_size(ext_gmp_type z)
{
  unsigned long r;
  if (z->_mp_size==0) return GMP_min;
  count_leading_zeros(r,(z->_mp_d)[abs(z->_mp_size)-1]);
  return (z->_mp_exp) *limb_size_in_bit - r;
}

void ext_gmp_add(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  GMP_SET_PREC(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : add set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpf_dump(z1);printf("\n");
    mpf_dump(z2);printf("\n");
  }
#endif
  mpf_add(z,z1,z2);
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpf_dump(z);printf("\n");
  }
#endif
  return;
}

void ext_gmp_add_i(ext_gmp_type z1,long n,ext_gmp_type z,long p)
{ 
	long q,s1,s2=limb_size_in_bit;
  	s1=ext_gmp_size(z1);
	q=max(s1,s2)-p+10;
	if (q<10)q=10;
	GMP_SET_PREC(z,q);
	#if DEBUG >0 
	#endif
	if(n<0)mpf_sub_ui(z,z1,-n);
		else mpf_add_ui(z,z1,n);
	#if (DEBUG >0)
	#endif
	return;
}


void ext_gmp_sub(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  GMP_SET_PREC(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : sub set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpf_dump(z1);printf("\n");
    mpf_dump(z2);printf("\n");
  }
#endif
  mpf_sub(z,z1,z2);
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpf_dump(z);printf("\n");
  }
#endif
  return;
}


void ext_gmp_sub_i(ext_gmp_type z1,long z2,ext_gmp_type z,long p)
{ 
	long q,s1,s2=limb_size_in_bit;
  	s1=ext_gmp_size(z1);
	q=max(s1,s2)-p+10;
  	if (q<10)q=10;
  	GMP_SET_PREC(z,q);
	#if DEBUG >0 
	#endif
  	if(z2<0)mpf_add_ui(z,z1,-z2); else mpf_sub_ui(z,z1,z2);
	#if (DEBUG >0)
	#endif
  	return;
}

void ext_gmp_i_sub(long z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ 
	long q,s1,s2=limb_size_in_bit;
  	s1=ext_gmp_size(z2);
	q=max(s1,s2)-p+10;
  	if (q<10)q=10;
  	GMP_SET_PREC(z,q);
	#if DEBUG >0 
	#endif
  	if(z1<0){mpf_add_ui(z,z2,-z1);mpf_neg(z,z);} else mpf_ui_sub(z,z1,z2);
	#if (DEBUG >0)
	#endif
  	return;
}



void ext_gmp_mul(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=s1+s2-p+10;
  if (q<10)q=10;
  GMP_SET_PREC(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : mul set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpf_dump(z1);printf("\n");
    mpf_dump(z2);printf("\n");
  }
#endif
  mpf_mul(z,z1,z2);
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpf_dump(z);printf("\n");
  }
#endif
  return;
}

void ext_gmp_mul_i(ext_gmp_type z1, long z2, ext_gmp_type z, long p)
{
long q,s1,zz,s2=limb_size_in_bit;
s1=ext_gmp_size(z1);
q=s1+s2-p+10;if(q<10)q=10;
GMP_SET_PREC(z,q);
#if DEBUG>0 
#endif
if (z2<0) zz=-z2; else zz=z2;
mpf_mul_ui(z,z1,zz);
#if (DEBUG>0)
#endif
if (z2<0)mpf_neg(z,z);
return;
}

void ext_gmp_div(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=s1-s2-p+10;
  if (q<10)q=10;
  GMP_SET_PREC(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : div set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpf_dump(z1);printf("\n");
    mpf_dump(z2);printf("\n");
  }
#endif
  mpf_div(z,z1,z2);
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpf_dump(z);printf("\n");
  }
#endif
  return;
}

void ext_gmp_div_i(ext_gmp_type z1,long z2,ext_gmp_type z,long p)
{ 
	long q,s1,zz,s2=1;
  	s1=ext_gmp_size(z1);
  	q=s1-s2-p+10;
  	if (q<10)q=10;
  	GMP_SET_PREC(z,q);
		#if DEBUG >0 
		#endif
	if(z2<0)zz=-z2;else zz=z2;
  	mpf_div_ui(z,z1,zz);
		#if (DEBUG >0)
		#endif
	if(z2<0)mpf_neg(z,z);
  	return;
}

void ext_gmp_i_div(long z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ 
	long q,s1=limb_size_in_bit,zz,s2;
  	s2=ext_gmp_size(z2);
  	q=s1-s2-p+10;
  	if (q<10)q=10;
  	GMP_SET_PREC(z,q);
		#if DEBUG >0 
		#endif
	if(z1<0)zz=-z1;else zz=z1;
  	mpf_ui_div(z,zz,z2);
		#if (DEBUG >0)
		#endif
	if(z1<0)mpf_neg(z,z);
  	return;
}




void ext_gmp_abs(ext_gmp_type z1,ext_gmp_type z)
{
  long q1=GMP_PREC(z1);
  GMP_SET_PREC(z,q1);
  mpf_abs(z,z1);
  return;
}

void ext_gmp_truncate(ext_gmp_type z1,ext_gmp_type z)
{
  long q1=GMP_PREC(z1);
  GMP_SET_PREC(z,q1);
  mpf_trunc(z,z1);
  return;
}


void ext_gmp_writee(ext_gmp_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpf_get_str (s, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) {
    if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
}

char* ext_gmp_swritee(ext_gmp_type z,long w)
{ char *s=(char *)malloc(w),*r;
  long e; 
  int i;
  int flag=0;
  r=mpf_get_str(NULL, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) {
    if ( r[i]==0 ) flag=1 ;
    if ( flag ) r[i]='0';
  } 
  if (r[0]=='-') {r[w-7]=0;sprintf(s,"-.%sE%+05ld",&r[1],e);}
  else {r[w-8]=0;sprintf(s,"+.%sE%+05ld",&r[0],e);};
  return s;
}

void ext_gmp_write(ext_gmp_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpf_get_str (s, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) { if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
  if (z->_mp_size != 0) {
     printf("\n");
     for (i=abs(z->_mp_size-1);i>=0; i-=1) 
       printf("%lu ",z->_mp_d[i]);
  }
  printf("\nExponent %ld, Limbs: %d, Bits: %ld\n",
     z->_mp_exp,z->_mp_size,GMP_PREC(z));
}

void ext_gmp_sqrt(ext_gmp_type z1,ext_gmp_type z,long p)
{ long q,s1;
  s1=ext_gmp_size(z1);
  q=s1/2-p+10;
  if (q<10)q=10;
  GMP_SET_PREC(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : sqrt set_prec: %ld %ld, p: %ld\n",op_count, s1,q,p);
  if  (op_count>minshowop){
    mpf_dump(z1);printf("\n");
  }
#endif
  mpf_sqrt(z,z1);
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpf_dump(z);printf("\n");
  }
#endif
  return;
}

void ext_gmp_shift(ext_gmp_type z1,ext_gmp_type z,long n)
{
  GMP_SET_PREC(z,GMP_PREC(z1));
  if ( n>=  0 ) mpf_mul_2exp(z,z1,n);
  else mpf_div_2exp(z,z1,-n);
  return;
}
void ext_gmp_set_z(ext_gmp_type r,int_gmp_type i)
{
  GMP_SET_PREC(r,mpz_sizeinbase(i,2));
  mpf_set_z(r,i);
}




