#include <stdio.h>

#ifndef DEFINITION_Dyadic_Base
#include "LRGMP_Dyadic_Base.h"
#endif

inline void Dyadic_Base_SysAdd
(Dyadic_Base_WertFeld A, Dyadic_Base_WertFeld B, Dyadic_Base_WertFeld C, long ANZAHL)
{
  if ( ANZAHL >= Dyadic_Base_MAXSIZE ) {
    printf("internal overflow in Dyadic_Base_SysAdd\n");
    exit(1);
  }
  C[ANZAHL]=mpn_add_n(C,A,B,ANZAHL);
}

inline void Dyadic_Base_SysSub
(Dyadic_Base_WertFeld A, Dyadic_Base_WertFeld B, Dyadic_Base_WertFeld C, long ANZAHL, BOOLEAN *NEGATIV)
{
  if ( ANZAHL >= Dyadic_Base_MAXSIZE ) {
    printf("internal overflow in Dyadic_Base_SysSub\n");
    exit(1);
  }
  if ( A[ANZAHL-1]> B[ANZAHL-1] ) {
    *NEGATIV=mpn_sub_n(C,A,B,ANZAHL);
  } else  if ( A[ANZAHL-1]< B[ANZAHL-1] ) {
    mpn_sub_n(C,B,A,ANZAHL);
    *NEGATIV=1;
  } else {
    *NEGATIV=mpn_sub_n(C,A,B,ANZAHL);
    if (*NEGATIV)  mpn_sub_n(C,B,A,ANZAHL);
  }
}


inline void Dyadic_Base_SysBlockShl
(Dyadic_Base_WertFeld A, long ANZAHL, long COUNT)
{
  if ( ANZAHL >= Dyadic_Base_MAXSIZE ) {
    printf("internal overflow in Dyadic_Base_SysBlockShl\n");
    exit(1);
  }
  if (COUNT != 0) mpn_lshift(A,A,ANZAHL,COUNT);
}

inline void Dyadic_Base_SysMul
(Dyadic_Base_WertFeld A, Dyadic_Base_WertFeld B, Dyadic_Base_WertFeld C, long Aanzahl, long Banzahl, long Canzahl)
{
/* Necessary condition: Aanzahl+Banzahl >= Canzahl */
  long ExpC;
  mp_limb_t *p,*q;
  if ( Aanzahl+Banzahl >= Dyadic_Base_MAXSIZE ) {
    printf("internal overflow in Dyadic_Base_SysMul\n");
    exit(1);
  }
  if ( Aanzahl == 1 )
    C[Banzahl]=mpn_mul_1(C,B,Banzahl,A[0]);
  else if ( Banzahl == 1 )
    C[Aanzahl]=mpn_mul_1(C,A,Aanzahl,B[0]);
  else if ( Aanzahl < Banzahl )
    mpn_mul(C,B,Banzahl,A,Aanzahl);
  else
    mpn_mul(C,A,Aanzahl,B,Banzahl);

  if (Aanzahl +Banzahl < Canzahl) {
    printf("parameters too big in Dyadic_Base_SysMul: %ld %ld\n",Aanzahl +Banzahl, Canzahl); 
    exit(1);
  } else if  (Aanzahl +Banzahl > Canzahl) {
/*	  printf("# %ld %ld\n",Aanzahl +Banzahl, Canzahl); */
     p=C;q=&C[Aanzahl+Banzahl-Canzahl];
     for (ExpC = 0;ExpC < Canzahl ; ExpC += 1) *(p++) = *(q++);
  }

}


inline void Dyadic_Base_SysDiv
(Dyadic_Base_WertFeld A, Dyadic_Base_WertFeld B, Dyadic_Base_WertFeld C, long Aanzahl, long Banzahl, long Canzahl, BOOLEAN MODUL)
{
/* Necessary conditon: CAnzahl = AAnzahl */
  long anzahl;
  unsigned long val,check,shiftcount;
  long ExpC;
  mp_limb_t *p,*q;
  anzahl=Banzahl+Canzahl-1;
  
  if ( Canzahl >= Dyadic_Base_MAXSIZE ) {
    printf("internal overflow in Dyadic_Base_SysDiv\n");
    exit(1);
  }
 /*if ( Canzahl != Aanzahl ) printf("%ld / %ld / %ld\n",Aanzahl,Banzahl,Canzahl);*/

  if ( Banzahl == 1) {
  A[0]=mpn_divrem_1(C,0,A,Aanzahl,B[0]);	 
  return;
  }	

  if (Banzahl> Aanzahl+1) {
    p=&A[Banzahl-1];q=&A[Aanzahl-1];
    for (ExpC = 0;ExpC < Aanzahl; ExpC += 1) *(p--) = *(q--); 
    for (ExpC = Aanzahl;ExpC < Banzahl; ExpC += 1) *(p--) = 0;
    Aanzahl=Banzahl;
  }
  A[Aanzahl]=0;
	 
  val=B[Banzahl-1];
  shiftcount=Dyadic_Base_BITZAHL-1; 
  for (check=Dyadic_Base_BITZAHL>>1; check>0; check=check>>1 ) {
    if (val>= (1<<check) ){
      shiftcount-=check; val=val>>check;
    }
  }

  if (shiftcount > 0 ) {
    mpn_lshift(B,B,Banzahl,shiftcount);
    A[Aanzahl]=mpn_lshift(A,A,Aanzahl,shiftcount);
  } 
  C[Aanzahl]=mpn_divrem(C,Banzahl-1,A,Aanzahl+1,B,Banzahl);	
  if (shiftcount > 0 ) {
    mpn_rshift(A,A,Banzahl,shiftcount);
  }

}

/*
inline void Dyadic_Base_SysMul_Test
(Dyadic_Base_WertFeld A, Dyadic_Base_WertFeld B, Dyadic_Base_WertFeld C, long Aanzahl, long Banzahl, long Canzahl)
{

  long ExpC;
  mp_limb_t *p,*q,carry;
  long a,b,c,size,m,d;
  if ( Aanzahl+Banzahl >= Dyadic_Base_MAXSIZE ) {
    printf("internal overflow in Dyadic_Base_SysMul\n");
    exit(1);
  }
  if (   Canzahl < 20 || Canzahl > 70 
      || Aanzahl < (Canzahl >> 1) || Banzahl < (Canzahl >> 1 ) ) {
    if ( Aanzahl < Banzahl )
      mpn_mul(C,B,Banzahl,A,Aanzahl);
    else
      mpn_mul(C,A,Aanzahl,B,Banzahl);
    if (Aanzahl +Banzahl < Canzahl) {
       p=&C[Canzahl-1];q=&C[Aanzahl+Banzahl-1];
       for (ExpC = 0;ExpC < Aanzahl+Banzahl; ExpC += 1) *(p--) = *(q--);
       p=&C[Canzahl-1-Aanzahl-Banzahl];
       for (ExpC = Aanzahl+Banzahl;ExpC < Canzahl; ExpC += 1) *(p--) = 0;
    } else if  (Aanzahl +Banzahl > Canzahl) {
       p=C;q=&C[Aanzahl+Banzahl-Canzahl];
       for (ExpC = 0;ExpC < Canzahl ; ExpC += 1) *(p++) = *(q++);
    }
    return;
  }
  for (c=0;c<=Canzahl;c++) C[c]=0;
  A[Aanzahl]=0;
  if (Aanzahl+Banzahl>Canzahl)C[0]=1;
  m=Banzahl-Canzahl-1; if (m<0) m=0;  
  d=0;
  for (b=m; b<Banzahl; b+=1) {
    a=Aanzahl+Banzahl-Canzahl-2-b;
    if (a<0)a=0;
    c=Canzahl-Aanzahl-Banzahl+a+b;
    size=Aanzahl-a;
    carry=mpn_addmul_1(&C[c],&A[a],size,B[b]);
    if ( carry ) mpn_add_1(&C[size+c],&C[size+c],d+1,carry);
    if ( C[size+c+d] != 0) d+=1;
    }
}
*/
