#include "MPFR_ext.h"
#include "gmp-mparam.h"
#include "mpfr-impl.h"

#define   min(a,b)        (a < b ? a : b)
#define   max(a,b)        (a < b ? b : a)
#define   max3(a,b,c)       max(a,max(b,c))
#define   max4(a,b,c,d)       max(max(a,b),max(c,d))

#ifndef DEBUG
#define DEBUG 0
#endif
#if DEBUG > 0
  long DEBUG_OP = 1;
  unsigned long op_count=0;
  long minshowop = DEBUG;
#endif

long ext_mpfr_var_count=0;

#define MaxFreeVars 1000
static ext_mpfr_type mpfr_FreeVars[MaxFreeVars];
static long mpfr_FreeVarCount=0L;

//#define MPFR_PREC(x) ((x)->_mpfr_prec)
//#define MPFR_MANT(x) ((x)->_mpfr_d)
#define MPFR_MSW_INDEX(x) (((x)->_mpfr_prec-1)/BITS_PER_MP_LIMB)

void ext_mpfr_remove_trailing_zeroes (mpfr_t x)
{
   unsigned int xn = MPFR_MSW_INDEX(x);
   unsigned int i;
   unsigned int j=(xn+1)*BITS_PER_MP_LIMB;
   mp_limb_t *xp = MPFR_MANT (x);
#if DEBUG >0 
  if  (op_count>minshowop){
    printf("pre-round: ");mpfr_print_raw(x);printf("\n"); 
  }
#endif
 
   if ( xp[0] != 0 ) return;
   /* first determine last non-zero limb */
   for (i=0; i<=xn && xp[i]==0; (i++) ) j-=BITS_PER_MP_LIMB;
 
   if (i > xn) return;
   mpfr_round_prec (x, GMP_RNDN, j); 
}                                                                                             


void ext_mpfr_initialize()
{
mpfr_set_default_prec(32);
}

ext_mpfr_type ext_mpfr_init()
{ ext_mpfr_type z;
  if (mpfr_FreeVarCount > 0) {
    mpfr_FreeVarCount -=1;
    z =mpfr_FreeVars[mpfr_FreeVarCount];
  } else {
    z = (ext_mpfr_type) malloc(sizeof (__mpfr_struct));
    mpfr_init(z);
  }
  ext_mpfr_var_count +=1;

// fprintf(stderr,"create %x\n",z);

  return z;
}

void ext_mpfr_free(ext_mpfr_type z)
{
//fprintf(stderr,"delete %x\n",z);
  if ( mpfr_FreeVarCount < MaxFreeVars ) {
    mpfr_FreeVars[mpfr_FreeVarCount]=z;
    mpfr_FreeVarCount +=1;
  } else {
    mpfr_clear(z); 
    free (z);
  }
  ext_mpfr_var_count -=1;
}

long ext_mpfr_size(ext_mpfr_type z)
{
if ( NOTZERO(z) ) return EXP(z);
else return GMP_min;
}

void ext_mpfr_getsize(ext_mpfr_type z,ext_mpfr_sizetype* s)
{
  int zn=MPFR_MSW_INDEX(z);
  if ( zn<0 ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else if ( ((z->_mpfr_d)[zn]==0) ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else {
    s->mantissa=((z->_mpfr_d)[zn])/2+1;
    s->exponent=EXP(z)-BITS_PER_MP_LIMB+1;
  }
//   printf("getsize: %lu*2^%ld\n",  s->mantissa,s->exponent);
}

void ext_mpfr_add(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=max(s1,s2)-p+1;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : add set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpfr_dump(z1,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_dump(z2,iRRAM_mpfr_rounding_mode);printf("\n\n");
    mpfr_print_raw(z1);printf("\n");
    mpfr_print_raw(z2);printf("\n"); 
  }
#endif
  mpfr_add(z,z1,z2,iRRAM_mpfr_rounding_mode);
  ext_mpfr_remove_trailing_zeroes (z); 
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_print_raw(z);printf("\n");
  }
#endif
  return;
}

void ext_mpfr_add_i(ext_mpfr_type z1,long n,ext_mpfr_type z,long p)
{ 
	long q,s1,s2=BITS_PER_MP_LIMB ;
  	s1=ext_mpfr_size(z1);
	q=max(s1,s2)-p+1;
	if (q<10)q=10;
	mpfr_set_prec(z,q);
	#if DEBUG >0 
	#endif
	if(n<0)mpfr_sub_ui(z,z1,-n,iRRAM_mpfr_rounding_mode);
		else mpfr_add_ui(z,z1,n,iRRAM_mpfr_rounding_mode);
	#if (DEBUG >0)
	#endif
	return;
}


void ext_mpfr_sub(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=max(s1,s2)-p+1;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : sub set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpfr_dump(z1,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_dump(z2,iRRAM_mpfr_rounding_mode);printf("\n\n");
    mpfr_print_raw(z1);printf("\n");
    mpfr_print_raw(z2);printf("\n"); 
  }
#endif
  mpfr_sub(z,z1,z2,iRRAM_mpfr_rounding_mode);
  ext_mpfr_remove_trailing_zeroes (z); 
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_print_raw(z);printf("\n");
  }
#endif
 return;
}

void ext_mpfr_sub_i(ext_mpfr_type z1,long n,ext_mpfr_type z,long p)
{ 
	long q,s1 ;
  	s1=ext_mpfr_size(z1);
	q=max(s1,BITS_PER_MP_LIMB)-p+1;
	if (q<10)q=10;
	mpfr_set_prec(z,q);
	#if DEBUG >0 
	#endif
	if(n<0)mpfr_add_ui(z,z1,-n,iRRAM_mpfr_rounding_mode);
		else mpfr_sub_ui(z,z1,n,iRRAM_mpfr_rounding_mode);
	#if (DEBUG >0)
	#endif
	return;
}

void ext_mpfr_i_sub(long z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ 
	long q,s2;
  	s2=ext_mpfr_size(z2);
	q=max(BITS_PER_MP_LIMB,s2)-p+1;
  	if (q<10)q=10;
  	mpfr_set_prec(z,q);
	#if DEBUG >0 
	#endif
  	if(z1<0){mpfr_add_ui(z,z2,-z1,iRRAM_mpfr_rounding_mode);mpfr_neg(z,z,iRRAM_mpfr_rounding_mode);} 
	else mpfr_ui_sub(z,z1,z2,iRRAM_mpfr_rounding_mode);
	#if (DEBUG >0)
	#endif
  	return;
}

void ext_mpfr_mul(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=s1+s2-p+1;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : mul set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpfr_dump(z1,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_dump(z2,iRRAM_mpfr_rounding_mode);printf("\n\n");
    mpfr_print_raw(z1);printf("\n");
    mpfr_print_raw(z2);printf("\n"); 
  }
#endif
  mpfr_mul(z,z1,z2,iRRAM_mpfr_rounding_mode);
  ext_mpfr_remove_trailing_zeroes (z); 
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_print_raw(z);printf("\n");
  }
#endif
  return;
}

void ext_mpfr_mul_i(ext_mpfr_type z1, long z2, ext_mpfr_type z, long p)
{
long q,s1,zz;
s1=ext_mpfr_size(z1);
q=s1+BITS_PER_MP_LIMB-p+10;if(q<10)q=10;
mpfr_set_prec(z,q);
#if DEBUG>0 
    printf("\n\n");mpfr_print_raw(z1);printf("\n");
    mpfr_dump(z1,iRRAM_mpfr_rounding_mode);;
    printf ("* %d = \n",z2);
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);;
    mpfr_print_raw(z);printf("\n");
#endif
if (z2<0) zz=-z2; else zz=z2;
mpfr_mul_ui(z,z1,zz,iRRAM_mpfr_rounding_mode);
#if (DEBUG>0)
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);;
    mpfr_print_raw(z);printf("\n\n");
#endif
if (z2<0)mpfr_neg(z,z,iRRAM_mpfr_rounding_mode);
return;
}

void ext_mpfr_div(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=s1-s2-p+1;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : div set_prec: %ld %ld %ld, p: %ld\n",op_count, s1,s2,q,p);
  if  (op_count>minshowop){
    mpfr_dump(z1,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_dump(z2,iRRAM_mpfr_rounding_mode);printf("\n\n");
    mpfr_print_raw(z1);printf("\n");
    mpfr_print_raw(z2);printf("\n"); 
  }
#endif
  mpfr_div(z,z1,z2,iRRAM_mpfr_rounding_mode);
  ext_mpfr_remove_trailing_zeroes (z); 
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_print_raw(z);printf("\n");
  }
#endif
  return;
}

void ext_mpfr_div_i(ext_mpfr_type z1,long z2,ext_mpfr_type z,long p)
{ 
	long q,s1,zz,s2=1;
  	s1=ext_mpfr_size(z1);
  	q=s1-s2-p+10;
  	if (q<10)q=10;
  	mpfr_set_prec(z,q);
	if(z2<0)zz=-z2;else zz=z2;
  	mpfr_div_ui(z,z1,zz,iRRAM_mpfr_rounding_mode);
	if(z2<0)mpfr_neg(z,z,iRRAM_mpfr_rounding_mode);
  	return;
}

void ext_mpfr_i_div(long z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ 
	long q,zz,s2;
  	s2=ext_mpfr_size(z2);
  	q=BITS_PER_MP_LIMB-s2-p+10;
  	if (q<10)q=10;
  	mpfr_set_prec(z,q);
		#if DEBUG >0 
		#endif
	if(z1<0)zz=-z1;else zz=z1;
  	mpfr_ui_div(z,zz,z2,iRRAM_mpfr_rounding_mode);
		#if (DEBUG >0)
		#endif
	if(z1<0)mpfr_neg(z,z,iRRAM_mpfr_rounding_mode);
  	return;
}

void ext_mpfr_abs(ext_mpfr_type z1,ext_mpfr_type z)
{
  long q1=mpfr_get_prec(z1);
  mpfr_set_prec(z,q1);
  mpfr_abs(z,z1,iRRAM_mpfr_rounding_mode);
  return;
}

void ext_mpfr_truncate(ext_mpfr_type z1,ext_mpfr_type z)
{
  long q1=mpfr_get_prec(z1);
  mpfr_set_prec(z,q1);
  mpfr_trunc(z,z1);
  return;
}


void ext_mpfr_writee(ext_mpfr_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpfr_get_str (s, &e, 10, w-7,z,iRRAM_mpfr_rounding_mode);
  for (i=0;i<w-7;i++) {
    if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
}

void ext_mpfr_write(ext_mpfr_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpfr_get_str (s, &e, 10, w-7,z,iRRAM_mpfr_rounding_mode);
  for (i=0;i<w-7;i++) { if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
  if (z->_mpfr_size != 0) {
     printf("\n");
     for (i=abs(z->_mpfr_size-1);i>=0; i-=1) 
       printf("%lu ",z->_mpfr_d[i]);
  }
  printf("\nExponent %ld, Limbs: %ld, Bits: %ld\n",
     z->_mpfr_exp,z->_mpfr_size,mpfr_get_prec(z));
}

void ext_mpfr_duplicate_wo_init(ext_mpfr_type z1,ext_mpfr_type z2)
{
  long q1=mpfr_get_prec(z1);
  if (mpfr_get_prec(z2)< q1) mpfr_set_prec(z2,q1);
  mpfr_set(z2,z1,iRRAM_mpfr_rounding_mode);
}

void ext_mpfr_duplicate_w_init(ext_mpfr_type z1,ext_mpfr_type *z2)
{
//  *z2 = (ext_mpfr_type) malloc(sizeof (__mpfr_struct));
//  mpfr_init2(*z2,mpfr_get_prec(z1));
  long q1=mpfr_get_prec(z1);
  *z2=ext_mpfr_init();
  if (mpfr_get_prec(*z2)< q1) mpfr_set_prec(*z2,q1);
//  ext_mpfr_var_count +=1;
  mpfr_set(*z2,z1,iRRAM_mpfr_rounding_mode);
}

void ext_mpfr_sqrt(ext_mpfr_type z1,ext_mpfr_type z,long p)
{ long q,s1;
  s1=ext_mpfr_size(z1);
  q=s1/2-p+1;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
#if DEBUG >0 
  op_count +=1;
  if (DEBUG_OP ||(op_count>minshowop))
  printf("op %ld : sqrt set_prec: %ld %ld, p: %ld\n",op_count,s1,q,p);
  if  (op_count>minshowop){
    mpfr_dump(z1,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_print_raw(z1);printf("\n");
   }
#endif
  mpfr_sqrt(z,z1,iRRAM_mpfr_rounding_mode);
  ext_mpfr_remove_trailing_zeroes (z); 
#if (DEBUG >0)
  if ( op_count > minshowop ) {
    mpfr_dump(z,iRRAM_mpfr_rounding_mode);printf("\n");
    mpfr_print_raw(z);printf("\n");
  }
#endif
  return;
}

void ext_mpfr_shift(ext_mpfr_type z1,ext_mpfr_type z,long n)
{
  mpfr_set_prec(z,mpfr_get_prec(z1));
  if ( n>=  0 ) mpfr_mul_2exp(z,z1,n,iRRAM_mpfr_rounding_mode);
  else mpfr_div_2exp(z,z1,-n,iRRAM_mpfr_rounding_mode);
  return;
}

void ext_mpfr_set_z(ext_mpfr_type r,int_mpfr_type i)
{
  long sib=mpz_sizeinbase(i,2);
  mpfr_set_prec(r,max(32,sib));
  mpfr_set_z(r,i,iRRAM_mpfr_rounding_mode);
} 

void ext_mpfr_set_f2i(ext_mpfr_type r,int_mpfr_type i){
  long exp=mpfr_get_z_exp (i,r);
  if (mpfr_sgn(r) < 0) {mpz_neg(i,i);}
  if ( exp > 0)  mpz_mul_2exp(i,i,exp); 
  else       mpz_tdiv_q_2exp(i,i,-exp);
};
