/*

limit.cc -- Templates for general limits, lipschitz limits ...
            for the iRRAM library
 
Copyright (C) 2001-2003 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


/*
Changelog: (initial version by Norbert)

  2001-09-11 correction of ordinary limits due to 
             moving from longjump to exception as iteration mechanism

*/

#include <math.h>
#include <stdlib.h>
#include <stdarg.h>

#include "iRRAM_core.h"

template <class ARGUMENT, class RESULT> 
RESULT limit (RESULT f(long prec,const ARGUMENT&),
                           const ARGUMENT& x)

{
  RESULT lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error;
  long element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  x.geterror(x_error);

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;
  ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

  DEBUG1(2,"starting limit_gen1...\n");
  while (1) {
    try {
    DEBUG2(2,"trying to compute limit_gen1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor 
    	&&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }
    if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
    }
    catch ( Iteration it) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}
  }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_gen1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}

template <class ARGUMENT, class RESULT> 
RESULT limit (RESULT f(long prec,const ARGUMENT&,long param),
                           const ARGUMENT& x,long param)

{
  RESULT lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error;
  long element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;
  ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

  x.geterror(x_error);
  DEBUG1(2,"starting limit_gen1...\n");
  while (1) {
   try { 
    DEBUG2(2,"trying to compute limit_gen1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x,param);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor 
                       &&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }
    if ( firsttime  != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
    }
    catch ( Iteration it) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}

  }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_gen1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REAL limit         (REAL f(long, const REAL&, const REAL&),
                           const REAL& x, 
                           const REAL& y)
{
  REAL lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error,y_error;
  long element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  x.geterror(x_error);
  y.geterror(y_error);

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;
  ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

  DEBUG1(2,"starting limit_gen2 ...\n");
  while (1) {
   try { 
    DEBUG2(2,"trying to compute limit_gen2 with precision %ld...\n",element);
    limnew=f(element,x,y);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor 
     &&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor
     &&  limnew_error.exponent > y_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }
    if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ){
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
    }
    catch ( Iteration it)  {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}
  }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_gen2 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REAL limit (REAL f(long))
{
  REAL lim;
  sizetype lim_error;

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;
  ACTUAL_STACK.prec_inc=ACTUAL_STACK.prec_inc+iRRAM_prec_inc2;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

  DEBUG1(2,"starting limit_0 ...\n");
  while (1) {
   try { 
    DEBUG2(2,"trying to compute limit_0 with precision %ld...\n",ACTUAL_STACK.actual_prec); 
    lim=f(SAVED_STACK.actual_prec);
    DEBUG3(2,"getting result with local error %ld*2^(%ld)\n", lim.error.mantissa, lim.error.exponent);
    break;
  }
    catch ( Iteration it) {
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_0 failed, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
   }}
  sizetype_set(lim_error,1,SAVED_STACK.actual_prec);
  lim.adderror(lim_error);
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_0 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL limit_lip (REAL f(long,const REAL&),
            long lip_value,
	    bool on_domain(const REAL&),
            const REAL& x)
{
  if ( on_domain(x) != true ) REITERATE(0);

  REAL x_new,lim;
  sizetype lim_error;

  x_new=x; 
  sizetype_exact(x_new.error);

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;
  ACTUAL_STACK.prec_inc=ACTUAL_STACK.prec_inc+iRRAM_prec_inc2;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

  DEBUG1(2,"starting limit_lip1 ...\n");
  while (1) {
     try{ 
      DEBUG2(2,"trying to compute limit_lip1 with precision %ld...\n",ACTUAL_STACK.actual_prec);
    lim=f(SAVED_STACK.actual_prec,x_new);
    if (lim.error.exponent > SAVED_STACK.actual_prec ) {
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip1 too imprecise, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    } else { 
      DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lim.error.mantissa, lim.error.exponent);
      break;
    }}
    catch ( Iteration it){
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip1 failed, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    } }
  sizetype_set(lim_error,1,SAVED_STACK.actual_prec);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,x.error,lip_value);
  lim.adderror(lim_error);
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_lip1 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            x.geterror(lim_error);
            fprintf(stderr,"  error of argument: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL limit_lip (REAL f(long,const REAL&),
            long lip_bound(const REAL&),
            const REAL& x)
{
  long lip_value; 
  continous_begin();
  lip_value=lip_bound(x);
  continous_end();

  ITERATION_STACK SAVED_STACK;

  REAL x_new,lim;
  sizetype lim_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  ACTUAL_STACK.inlimit+=1;
  DEBUG1(2,"starting limit_lip1 ...\n");
  ACTUAL_STACK.prec_inc=ACTUAL_STACK.prec_inc+iRRAM_prec_inc2;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
  while (1) {
     try{ 
      DEBUG2(2,"trying to compute limit_lip1 with precision %ld...\n",ACTUAL_STACK.actual_prec);
    lim=f(SAVED_STACK.actual_prec,x_new);
    if (lim.error.exponent > SAVED_STACK.actual_prec ) {
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip1 too imprecise, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    } else { 
      DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lim.error.mantissa, lim.error.exponent);
      break;
    }}
    catch ( Iteration it){
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip1 failed, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    } }
  sizetype_set(lim_error,1,SAVED_STACK.actual_prec);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,x.error,lip_value);
  lim.adderror(lim_error);
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_lip1 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            x.geterror(lim_error);
            fprintf(stderr,"  error of argument: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL limit_lip     (REAL f(long, const REAL&, const REAL&),
         long lip,
         bool on_domain(const REAL&,const REAL&),
         const REAL& x,
         const REAL& y)
{
  if ( on_domain(x,y) != true ) REITERATE(0);

  ITERATION_STACK SAVED_STACK;
  REAL x_new,y_new,lim;
  sizetype lim_error;

  x_new=x;   sizetype_exact(x_new.error);
  y_new=y;   sizetype_exact(y_new.error);
  ACTUAL_STACK.inlimit+=1;
  DEBUG1(2,"starting limit_lip2 ...\n");
  ACTUAL_STACK.prec_inc=ACTUAL_STACK.prec_inc+iRRAM_prec_inc2;
  while (1) {
    try { 
      DEBUG2(2,"trying to compute limit_lip2 with precision %ld...\n",ACTUAL_STACK.actual_prec);
      lim=f(SAVED_STACK.actual_prec,x_new,y_new);
     if (lim.error.exponent > SAVED_STACK.actual_prec ) {
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip2 too imprecise, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    } else { 
      DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lim.error.mantissa, lim.error.exponent);
      break;
    }}
    catch ( Iteration it) {
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip2 failed, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    }}
  sizetype_set(lim_error,1,SAVED_STACK.actual_prec);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,x.error,lip);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,y.error,lip);
  lim.adderror(lim_error);
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_lip2 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            x.geterror(lim_error);
            fprintf(stderr,"  error of argument 1: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            y.geterror(lim_error);
            fprintf(stderr,"  error of argument 2: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL lipschitz (REAL f(const REAL&),
            long lip,
	    bool on_domain(const REAL&),
            const REAL& x)
{
  if ( on_domain(x) != true ) REITERATE(0);

  REAL x_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz1 ...\n");
  lip_result=f(x_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_1 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  for argument with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(const REAL&),
            REAL lip_f(const REAL&),
	    bool on_domain(const REAL&),
            const REAL& x)
{
  if ( on_domain(x) != true ) REITERATE(0);

  REAL x_new,lip_result,lip_bound;
  sizetype lip_error,lip_size,tmp_size;

  x_new=x; 
  sizetype_exact(x_new.error);
  ITERATION_STACK SAVED_STACK;

  continous_begin ();
  DEBUG1(2,"starting lipschitz1b ...\n");
  lip_bound=lip_f(x);

  bool try_it=true;
  while (try_it) { 
  try { try_it=false;
        lip_result=f(x_new); }
  catch ( Iteration it)  { try_it=true;  
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip2 failed, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    }
  }
  continous_end ();

  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  MP_getsize(lip_bound.value,lip_size);
  lip_bound.geterror(tmp_size);
  sizetype_inc(lip_size,tmp_size);
  x.geterror(tmp_size);
  sizetype_mult(lip_error,lip_size,tmp_size);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_1b with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  for argument with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(const REAL&),
            REAL lip_f(const REAL&),
            const REAL& x)
{
  REAL x_new,lip_result,lip_bound;
  sizetype lip_error,lip_size,tmp_size;

  x_new=x; 
  sizetype_exact(x_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz1a ...\n");
  lip_result=f(x_new);
  lip_bound=lip_f(x);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  MP_getsize(lip_bound.value,lip_size);
  lip_bound.geterror(tmp_size);
  sizetype_inc(lip_size,tmp_size);
  x.geterror(tmp_size);
  sizetype_mult(lip_error,lip_size,tmp_size);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_1a with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  for argument with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(long, const REAL&),
            long lip,
	    bool on_domain(long k,const REAL&),
            long k,
            const REAL& x)
{
  if ( on_domain(k,x) != true ) REITERATE(0);
  REAL x_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz1 ...\n");
  lip_result=f(k,x_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_1 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  error of argument: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(const REAL&, const REAL&),
            long lip,
	    bool on_domain(const REAL&,const REAL&),
            const REAL& x,
            const REAL& y)
{
  if ( on_domain(x,y) != true ) REITERATE(0);
  REAL x_new,y_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  y_new=y;
  sizetype_exact(y_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz2 ...\n");
  lip_result=f(x_new,y_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  sizetype_shift(lip_error,y.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_2 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  error of argument x: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            y.geterror(lip_error);
            fprintf(stderr,"  error of argument y: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(long, const REAL&, const REAL&),
            long lip,
	    bool on_domain(long k,const REAL&,const REAL&),
            long k,
            const REAL& x,
            const REAL& y)
{
  if ( on_domain(k,x,y) != true ) REITERATE(0);
  REAL x_new,y_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  y_new=y;
  sizetype_exact(y_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz2 ...\n");
  lip_result=f(k,x_new,y_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  sizetype_shift(lip_error,y.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_2 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  error of argument x: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            y.geterror(lip_error);
            fprintf(stderr,"  error of argument y: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL limit_hint    (REAL f(long, const REAL&),
                    long hint,
                    const REAL& x)
{
  ITERATION_STACK LOCAL_STACK;
  REAL lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error;
  long hintcopy=hint;
  int success=0;

  x.geterror(element_error);
  long element=max(element_error.exponent,ACTUAL_STACK.actual_prec);

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;
  DEBUG1(2,"starting limit_hint1...\n");

  while (1) {
    try {
    DEBUG2(2,"trying to compute limit_hint1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if ( (! success) || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      hintcopy=2*hintcopy;
      }
    success=1; 
    if (element>iRRAM_starting_prec)break;
    element=element+hint; 
  }
    catch ( Iteration it) {
    if (element>iRRAM_starting_prec)break;
    element=element+hintcopy; 
    }}
  if ( ! success) {
    DEBUG1(1,"computation of limit_hint1 failed\n");
    REITERATE(0);
  }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_hint1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REAL limit_hint    (REAL f(long, const REAL&, const REAL&),
                    long hint,
                    const REAL& x, const REAL&y)
{
  ITERATION_STACK LOCAL_STACK;
  REAL lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error;
  long hintcopy=hint;
  int success=0;

  x.geterror(element_error);
  long element=max(element_error.exponent,ACTUAL_STACK.actual_prec);

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;

  DEBUG1(2,"starting limit_hint1...\n");
  while (1) {
    try {
    DEBUG2(2,"trying to compute limit_hint1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x,y);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if ( (! success) || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      hintcopy=2*hintcopy;
      }
    success=1; 
    if (element>iRRAM_starting_prec)break;
    element=element+hint; 
  }
    catch ( Iteration it) {
    if (element>iRRAM_starting_prec)break;
    element=element+hintcopy; 
    }}
  if ( ! success) {
    DEBUG1(1,"computation of limit_hint1 failed\n");
    REITERATE(0);
  }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_hint1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REALMATRIX limit_lip (REALMATRIX f(long,const REALMATRIX&),
            long lip,
	    bool on_domain(const REALMATRIX&),
            const REALMATRIX& x)
{
  if ( on_domain(x) != true ) REITERATE(0);
  REALMATRIX x_new,lim,limnew;
  sizetype lim_error;

  sizetype_exact(lim_error);
  x_new=x; x_new.seterror(lim_error);

  ITERATION_STACK SAVED_STACK;
  ACTUAL_STACK.inlimit+=1;

  ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
  ACTUAL_STACK.prec_inc=ACTUAL_STACK.prec_inc+iRRAM_prec_inc2;

  DEBUG1(2,"starting limit_matrix_lip1 ...\n");
  while (1) {
   try { 
      DEBUG2(2,"trying to compute limit_matrix_lip1 with precision %ld...\n",ACTUAL_STACK.actual_prec);
 
    limnew=f(SAVED_STACK.actual_prec,x_new);
    ACTUAL_STACK=SAVED_STACK;
    lim=limnew;
    DEBUG0(2,{lim.geterror(lim_error);
              fprintf(stderr,"getting result with local error %ld*2^(%ld)\n",
                lim_error.mantissa,lim_error.exponent);});
    break;
  }
    catch ( Iteration it)  {
      ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
      ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
      DEBUG2(2,"limit_lip1 failed, increasing precision locally to %ld...\n",ACTUAL_STACK.actual_prec);
    }}
  sizetype_exact(lim_error);
  lim.adderror(lim_error);
  x.geterror(lim_error);
  sizetype_shift(lim_error,lim_error,lip);
  lim.adderror(lim_error);
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_matrix_lip1 with error %ld*2^(%ld)\n",
                lim_error.mantissa,lim_error.exponent);});

  return lim;
}



template <class ARGUMENT, class RESULT> 
RESULT limit_mv (RESULT f(long prec,
                          long* choice,
                          const ARGUMENT&),
                          const ARGUMENT& x)
{
  RESULT lim,limnew;
  sizetype limnew_error,element_error;
  long choice=0;
  sizetype lim_error,x_error;
  long element=ACTUAL_STACK.actual_prec;
  int firsttime=2;

  if ( ! ACTUAL_STACK.inlimit ) {
	if (testanzahl > testaufrufe)
	{ 
		testaufrufe++;
		choice=testwert[testaufrufe];
	}
	else
	{
		cacheinc_test();
		testaufrufe++;
		testwert[testaufrufe]=0;
	}
  }

  ITERATION_STACK SAVED_STACK;

  ACTUAL_STACK.inlimit+=1;

  ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
  ACTUAL_STACK.actual_prec=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 

  x.geterror(x_error);
  DEBUG1(2,"starting limit_mv...\n");
  while (1) {
   try { 
    DEBUG2(2,"trying to compute limit_mv with precicion 2^(%ld)...\n",element);
    limnew=f(element,&choice,x);
    if ( ! SAVED_STACK.inlimit) testwert[testaufrufe]=choice;
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > SAVED_STACK.actual_prec -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor &&  limnew_error.exponent > x_error.exponent -SAVED_STACK.prec_inc/ACTUAL_STACK.prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+iRRAM_starting_prec;
       ACTUAL_STACK.prec_inc=0;
       firsttime=1;
    }}
     catch ( Iteration it) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+iRRAM_starting_prec;
      ACTUAL_STACK.prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      REITERATE(0);
      }}
   if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=SAVED_STACK.actual_prec)break;
    ACTUAL_STACK.prec_inc=long(ACTUAL_STACK.prec_inc * ACTUAL_STACK.prec_factor)+iRRAM_prec_inc1;
    element=iRRAM_starting_prec+ACTUAL_STACK.prec_inc; 
    }
  lim.seterror(lim_error);
  DEBUG0(2,{fprintf(stderr,"end of limit_mv with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


// Instantiation of templates for multi-valued limits:

template COMPLEX limit_mv<COMPLEX,COMPLEX>
    (COMPLEX (*)(long, long *, COMPLEX const &), COMPLEX const &);

template REALMATRIX limit_mv<REALMATRIX,REALMATRIX>
    (REALMATRIX (*)(long, long *, REALMATRIX const &), REALMATRIX const &);

// Instantiation of templates for single-valued limits, one parameter:

template REALMATRIX limit <REALMATRIX,REALMATRIX>
    (REALMATRIX (*)(long, REALMATRIX const &), REALMATRIX const &);

template REAL limit <REAL,REAL>
    (REAL (*)(long, REAL const &), REAL const &);

template COMPLEX limit <COMPLEX,COMPLEX>
    (COMPLEX (*)(long, COMPLEX const &), COMPLEX const &);

// Instantiation of templates for single-valued limits, one parameter + long:

template REAL limit <REAL,REAL>
    (REAL (*)(long, REAL const &, long param), REAL const &,long param);
